header {* \section{Simplest Examples} *}

theory OG_Examples_Simple imports "~~/src/HOL/Hoare_Parallel/OG_Syntax"

begin

record very_simple_1 =
 x :: nat


(* Purely sequential problem *)
lemma very_simple_1: 
  "\<parallel>- .{\<acute>x=0 }.  
  COBEGIN 
  .{\<acute>x=0 }.
  \<acute>x:=5;;
  .{ \<acute>x=5 }.  
  \<acute>x:=\<acute>x+1
  .{ \<acute>x=6 }.  
  COEND
  .{ \<acute>x=6 }."
apply oghoare
apply auto
done


lemma very_simple_2: 
  "\<parallel>- .{\<acute>x=0 }.  
  COBEGIN
  .{ True }.
  \<acute>x:=0
  .{ \<acute>x=0 \<or> \<acute>x=2  }.
  \<parallel>
  .{\<acute>x=0 }.
  \<acute>x:=\<acute>x+2
  .{ \<acute>x=0 \<or> \<acute>x=2  }.
  COEND
  .{ \<acute>x=0 \<or> \<acute>x=2  }."
apply oghoare
apply auto
done


subsection {* Mutual Exclusion *}

subsubsection {* Peterson's Algorithm I*}

text {* Eike Best. "Semantics of Sequential and Parallel Programs", page 217. *}

record Petersons_mutex_1 =
 pr1 :: nat
 pr2 :: nat
 in1 :: bool
 in2 :: bool 
 hold :: nat


lemma Petersons_mutex_1: 
  "\<parallel>- .{\<acute>pr1=0 \<and> \<not>\<acute>in1 \<and> \<acute>pr2=0 \<and> \<not>\<acute>in2 }.  
  COBEGIN 
  .{\<acute>pr1=0 \<and> \<not>\<acute>in1 }.  
  WHILE True INV .{\<acute>pr1=0 \<and> \<not>\<acute>in1 }.  
  DO  
  .{\<acute>pr1=0 \<and> \<not>\<acute>in1}. \<langle> \<acute>in1:=True,,\<acute>pr1:=1 \<rangle>;;  
  .{\<acute>pr1=1 \<and> \<acute>in1}.  \<langle> \<acute>hold:=1,,\<acute>pr1:=2 \<rangle>;;  
  .{\<acute>pr1=2 \<and> \<acute>in1 \<and> (\<acute>hold=1 \<or> \<acute>hold=2 \<and> \<acute>pr2=2)}.
  AWAIT (\<not>\<acute>in2 \<or> \<not>(\<acute>hold=1)) THEN \<acute>pr1:=3 END;;    
  .{\<acute>pr1=3 \<and> \<acute>in1 \<and> (\<acute>hold=1 \<or> \<acute>hold=2 \<and> \<acute>pr2=2) }. 
   \<langle>\<acute>in1:=False,,\<acute>pr1:=0\<rangle> 
  OD .{\<acute>pr1=0 \<and> \<not>\<acute>in1}.  
  \<parallel>  
  .{\<acute>pr2=0 \<and> \<not>\<acute>in2}.  
  WHILE True INV .{\<acute>pr2=0 \<and> \<not>\<acute>in2}.  
  DO  
  .{\<acute>pr2=0 \<and> \<not>\<acute>in2}. \<langle> \<acute>in2:=True,,\<acute>pr2:=1 \<rangle>;;  
  .{\<acute>pr2=1 \<and> \<acute>in2}. \<langle>  \<acute>hold:=2,,\<acute>pr2:=2 \<rangle>;;  
  .{\<acute>pr2=2 \<and> \<acute>in2 \<and> (\<acute>hold=2 \<or> (\<acute>hold=1 \<and> \<acute>pr1=2))}.  
  AWAIT (\<not>\<acute>in1 \<or> \<not>(\<acute>hold=2)) THEN \<acute>pr2:=3  END;;    
  .{\<acute>pr2=3 \<and> \<acute>in2 \<and> (\<acute>hold=2 \<or> (\<acute>hold=1 \<and> \<acute>pr1=2))}. 
    \<langle>\<acute>in2:=False,,\<acute>pr2:=0\<rangle> 
  OD .{\<acute>pr2=0 \<and> \<not>\<acute>in2}.  
  COEND  
  .{\<acute>pr1=0 \<and> \<not>\<acute>in1 \<and> \<acute>pr2=0 \<and> \<not>\<acute>in2}."
apply oghoare
--{* 104 verification conditions. *}
apply auto
done

end
